#define _POSIX_C_SOURCE 200809L

#include <lektor/common.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <pthread.h>

/* Abort function */
void
__lkt_abort(void)
{
    LOG_ERROR("ABORT", "Got aborted, I will provoque a segv to try to display the backtrace...");
    int *segv = (int *) - 1;
    *segv = 1;
}

/* Assert function */

static void
__no_assert(const char UNUSED *f, int UNUSED l, const char UNUSED *fn, const char UNUSED *m)
{
}

static void
__yes_assert(const char *file, int line, const char *func, const char *msg)
{
    LOG_ERROR("ASSERT", "Assertion failed: %s:%s:%d: %s", file, func, line, msg);
    LOG_ERROR("ASSERT", "I will provoque a segv to try to display the backtrace...");
    int *segv = (int *) - 1;
    *segv = 1;
}

void (*__lkt_assert)(const char *file, int line, const char *func, const char *msg) = NULL;

__attribute__((constructor)) static void
__set_assert(void)
{
    const char *no_assert = getenv(LKT_ENV_SET_ASSERT);
    if (NULL == no_assert || strlen(no_assert) == 0)
        __lkt_assert = __yes_assert;
    else
        __lkt_assert = __no_assert;
}

/* Log functions */

enum log_level __log_level = __LAST_UNUSED_LOG_LEVEL; /* All by default */

void
__lkt_log(enum log_level level, const char *section, const char *func, const char *file,
          uint64_t line_number, const char *format, ...)
{
    char line[LKT_MESSAGE_MAX];
    va_list ap;
    int hour = 0, min = 0, sec = 0;
    time_t rawtime = time(NULL);
    struct tm *timeval = localtime(&rawtime);
    if (timeval != NULL) {
        hour = timeval->tm_hour;
        min  = timeval->tm_min;
        sec  = timeval->tm_sec;
    }

    size_t skip_dir = 0;
    do {
        file += skip_dir + 1;
        skip_dir = strcspn(file, "/");
    } while (file[skip_dir] && file[skip_dir + 1]);

    if (level > __log_level)
        return;

    safe_snprintf(line, LKT_MESSAGE_MAX, " %c [%02d:%02d:%02d] %-10s %s(%s+%ld): %s\n",
                  level == ERROR ? '!' : level == WARN ? '*' : level == INFO ? '.' : ' ',
                  hour, min, sec,
                  section, func, file, line_number, format);

    va_start(ap, format);
    vfprintf(stderr, line, ap);
    va_end(ap);
}

/* Not implemented */

void
__not_implemented(const char *func, char *file, int line)
{
    LOG_ERROR("GENERAL", "Function %s in file %s at line %d not "
              "implemented", func, file, line);
    abort();
}

/* Endian stuff */

float
be_float_t(const uint8_t bytes[])
{
    union {
        uint32_t _uint;
        float _float;
    } ret;
    ret._uint = be_uint32_t(bytes, 4);
    return ret._float;
}

uint64_t
be_uint64_t(const uint8_t bytes[], size_t n)
{
    uint64_t res = 0;
    for (size_t i = 0; i < n; i++) {
#if defined (LKT_ENDIANES_LITTLE) && (LKT_ENDIANES_LITTLE == 1)
        res = (res << 8u) | bytes[i];
#else
        res = (res >> 8u) | bytes[i];
#endif
    }
    return res;
}

double
be_double_t(const uint8_t bytes[])
{
    union {
        uint64_t _uint;
        double _double;
    } ret;
    ret._uint = be_uint64_t(bytes, 8);
    return ret._double;
}

uint32_t
be_uint32_t(const uint8_t bytes[], size_t n)
{
    uint32_t res = 0;
    for (size_t i = 0; i < n; i++) {
#if defined (LKT_ENDIANES_LITTLE) && (LKT_ENDIANES_LITTLE == 1)
        res = (res << 8u) | bytes[i];
#else
        res = (res >> 8u) | bytes[i];
#endif
    }
    return res;
}

/* End of endian stuff */

char *
trim(char *str, size_t len, char c)
{
    char *res = (char *) str;
    char *end;

    for (; len != 0 && *res == c; ++str, --len)
        continue;

    if (*res == 0)
        return res;

    end = res + len - sizeof(char);
    len = 0;

    for (; res < end && *end == c; --end, ++len)
        continue;

    if (len > 0)
        end[1] = 0;

    return res;
}

int
is_utf8(const char *string)
{
    if (!string)
        return 1;

    const unsigned char *bytes = (const unsigned char *)string;
    while (*bytes) {
        /* ASCII */
        if  (bytes[0] == 0x09 || bytes[0] == 0x0A ||
             bytes[0] == 0x0D || (0x20 <= bytes[0] && bytes[0] <= 0x7E)) {
            bytes += 1;
            continue;
        }

        /* non-overlong 2-byte */
        if ((0xC2 <= bytes[0] && bytes[0] <= 0xDF) &&
            (0x80 <= bytes[1] && bytes[1] <= 0xBF)) {
            bytes += 2;
            continue;
        }

        if ( (bytes[0] == 0xE0 &&
              (0xA0 <= bytes[1] && bytes[1] <= 0xBF)  &&
              (0x80 <= bytes[2] && bytes[2] <= 0xBF)) ||    /* excluding overlongs */
             (((0xE1 <= bytes[0] && bytes[0] <= 0xEC) ||
               bytes[0] == 0xEE || bytes[0] == 0xEF)  &&
              (0x80 <= bytes[1] && bytes[1] <= 0xBF)  &&
              (0x80 <= bytes[2] && bytes[2] <= 0xBF)) ||    /* straight 3-byte */
             (bytes[0] == 0xED &&
              (0x80 <= bytes[1] && bytes[1] <= 0x9F)  &&
              (0x80 <= bytes[2] && bytes[2] <= 0xBF))) {    /* excluding surrogates */
            bytes += 3;
            continue;
        }

        if ( (bytes[0] == 0xF0  &&                          /* planes 1-3 */
              (0x90 <= bytes[1] && bytes[1] <= 0xBF)  &&
              (0x80 <= bytes[2] && bytes[2] <= 0xBF)  &&
              (0x80 <= bytes[3] && bytes[3] <= 0xBF)) ||
             ((0xF1 <= bytes[0] && bytes[0] <= 0xF3)  &&
              (0x80 <= bytes[1] && bytes[1] <= 0xBF)  &&
              (0x80 <= bytes[2] && bytes[2] <= 0xBF)  &&
              (0x80 <= bytes[3] && bytes[3] <= 0xBF)) ||    /* planes 4-15 */
             (bytes[0] == 0xF4 &&
              (0x80 <= bytes[1] && bytes[1] <= 0x8F)  &&
              (0x80 <= bytes[2] && bytes[2] <= 0xBF)  &&
              (0x80 <= bytes[3] && bytes[3] <= 0xBF))) {    /* plane 16 */
            bytes += 4;
            continue;
        }

        return 2;
    }

    return 0;
}

int
get_stdin_line(const char *prompt, char *buf, size_t len)
{
    if (prompt)
        fprintf(stdout, "%s", prompt);
    if (!fgets(buf, len, stdin))
        return 1;
    buf[len - 1] = 0u;
    return is_utf8(buf);
}

inline long
get_mtime(const char *path)
{
    struct stat statbuf = {0};
    return (stat(path, &statbuf) == -1) ? 0 : statbuf.st_mtime;
}

void *
safe_malloc(size_t size)
{
    void *tmp;
    tmp = malloc(size);
    if (!tmp) {
        LOG_ERROR("GENERAL", "Out of memory, can't alloc "
                  "'%ld' Bytes", size);
        exit(1);
    }
    return tmp;
}

void *
safe_zero_malloc(size_t size)
{
    void *ret = safe_malloc(size);
    memset(ret, 0, size);
    return ret;
}

int
read_self_exe(char *path, size_t len)
{
    return ! ((readlink(SELF_EXECUTABLE_LINUX,   path, len - 1) > 0)
              || (readlink(SELF_EXECUTABLE_FREEBSD, path, len - 1) > 0)
              || (readlink(SELF_EXECUTABLE_SOLARIS, path, len - 1) > 0));
}

int
safe_snprintf(char *dest, const size_t max_len, const char *format, ...)
{
    va_list ap;
    va_start(ap, format);
    int ret = vsnprintf(dest, max_len, format, ap);
    dest[max_len - 1] = '\0';
    va_end(ap);
    return ret;
}

long
long_length(long integer)
{
    long count = 0;
    while (integer) {
        integer /= 10;
        ++count;
    }
    return count;
}

char *
safe_strncpy(char *dest, const char *src, size_t n)
{
    char *ret = strncpy(dest, src, n);
    dest[n - 1] = '\0';
    return ret;
}

int
iter_string(const char *str_list, const char *sep, char *dest, const size_t dest_len, size_t *save)
{
    size_t len = strcspn(&str_list[*save], sep);

    if (len == 0)
        return 1;
    if (len >= dest_len)
        return -1;

    strncpy(dest, &str_list[*save], len);
    dest[len] = '\0';
    *save += len + 1;
    return 0;
}
